using System;
using System.Text;

namespace Smdn.Protocols.Imap4 {
  public static class ImapConverter
  {
    public static string FromModifiedUTF7String(string str)
    {
      // RFC 3501 INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1
      // 5.1.3. Mailbox International Naming Convention
      // http://tools.ietf.org/html/rfc3501#section-5.1.3
      if (!str.Contains("&"))
        return str;

      var bytes = Encoding.ASCII.GetBytes(str);
      var decoded = new StringBuilder();

      for (var index = 0; index < bytes.Length; index++) {
        // In modified UTF-7, printable US-ASCII characters, except for "&",
        // represent themselves
        // "&" is used to shift to modified BASE64
        if (bytes[index] != 0x26) { // '&'
          decoded.Append((char)bytes[index]);
          continue;
        }

        if (bytes.Length <= ++index)
          // incorrect form
          break;

        if (bytes[index] == 0x2d) { // '-'
          // The character "&" (0x26) is represented by the two-octet sequence "&-".
          decoded.Append('&');
          continue;
        }

        var nonprintable = new StringBuilder();

        for (; index < bytes.Length; index++) {
          if (bytes[index] == 0x2d) // '-'
            // "-" is used to shift back to US-ASCII
            break;

          nonprintable.Append((char)bytes[index]);
        }

        // modified UTF7 -> string
        decoded.Append(FromModifiedUTF7(nonprintable.ToString()));
      }

      return decoded.ToString();
    }

    private static string FromModifiedUTF7(string str)
    {
      return Encoding.BigEndianUnicode.GetString(FromModifiedBase64(str));
    }

    private static byte[] FromModifiedBase64(string str)
    {
      // "," is used instead of "/"
      str = str.Replace(',', '/');

      var padding = 4 - str.Length % 4;

      if (padding == 0)
        return Convert.FromBase64String(str);
      else
        return Convert.FromBase64String(str + (new string('=', padding)));
    }

    public static string ToModifiedUTF7String(string str)
    {
      // RFC 3501 INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1
      // 5.1.3. Mailbox International Naming Convention
      // http://tools.ietf.org/html/rfc3501#section-5.1.3

      var encoded = new StringBuilder();
      var shiftFrom = -1;

      for (var index = 0; index < str.Length; index++) {
        var c = str[index];

        if ((0x20 <= c && c <= 0x7e)) {
          if (0 <= shiftFrom) {
            // string -> modified UTF7
            encoded.Append('&');
            encoded.Append(ToModifiedUTF7(str.Substring(shiftFrom, index - shiftFrom)));
            encoded.Append('-');

            shiftFrom = -1;
          }

          // printable US-ASCII characters
          if (c == 0x26)
            // except for "&"
            encoded.Append("&-");
          else
            encoded.Append(c);
        }
        else {
          if (shiftFrom == -1)
            shiftFrom = index;
        }
      }

      if (0 <= shiftFrom) {
        // string -> modified UTF7
        encoded.Append('&');
        encoded.Append(ToModifiedUTF7(str.Substring(shiftFrom)));
        encoded.Append('-');
      }

      return encoded.ToString();
    }

    private static string ToModifiedUTF7(string str)
    {
      return ToModifiedBase64(Encoding.BigEndianUnicode.GetBytes(str));
    }

    private static string ToModifiedBase64(byte[] bytes)
    {
      var base64 = Convert.ToBase64String(bytes).Replace('/', ',');

      if (base64.EndsWith("="))
        base64 = base64.Substring(0, base64.IndexOf('='));

      return base64;
    }

    internal static ImapAuthenticationType ToAuthType(string authType)
    {
      switch (authType.ToUpper()) {
        case "LOGIN":
          return ImapAuthenticationType.Login;
        case "PLAIN":
          return ImapAuthenticationType.Plain;
        case "CRAM-MD5":
          return ImapAuthenticationType.CramMd5;
        case "KERBEROS_V4":
          return ImapAuthenticationType.KerberosV4;
        case "*":
          return ImapAuthenticationType.Any;
        default:
          throw new NotSupportedException("unsupported authentication type");
      }
    }

    internal static string ToString(ImapAuthenticationType authType)
    {
      switch (authType) {
        case ImapAuthenticationType.Login:
          return "login";
        case ImapAuthenticationType.Plain:
          return "plain";
        case ImapAuthenticationType.CramMd5:
          return "cram-md5";
        case ImapAuthenticationType.KerberosV4:
          return "kerberos_v4";
        case ImapAuthenticationType.Any:
          return "*";
        case ImapAuthenticationType.None:
          return string.Empty;
        default:
          throw new ArgumentOutOfRangeException();
      }
    }

    internal static ImapListType ToListType(string listType)
    {
      switch (listType.ToUpper()) {
        case "LIST":
          return ImapListType.List;
        case "LSUB":
          return ImapListType.Lsub;
        default:
          throw new FormatException("invalid list type, it must be LIST or LSUB");
      }
    }

    internal static string ToString(ImapListType listType)
    {
      switch (listType) {
        case ImapListType.List:
          return "list";
        case ImapListType.Lsub:
          return "lsub";
        default:
          throw new ArgumentOutOfRangeException();
      }
    }
  }
}
