using System;
using System.Collections.Generic;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;

using Smdn.Protocols.Pop3.Session;

namespace Smdn.Protocols.Pop3.Client {
  // Layers:
  // * Client.PopClient
  //   Client.PopClientAgent
  //   Session.PopSessionAgent
  //   Session.PopSession
  //   Transaction.PopTransactionAgent
  //   Transaction.PopTransaction/Transaction.PopRequest/Transaction.PopResponse
  //   Protocol.PopConnection/Protocol.PopProtocolCommand/Protocol.PopProtocolResponse
  //   (POP3)

  public class PopClient {
    protected PopClientAgent Agent {
      get { return agent; }
    }

    public PopSession Session {
      get { return session; }
      internal set
      {
        if (session != value) {
          session = value;
          if (session != null)
            session.Timeout = timeout;
        }
      }
    }

    public PopUri HostUri {
      get
      {
        if (session == null)
          return hostUri;
        else
          return session.HostUri;
      }
      set
      {
        if (session == null) {
          if (value == null)
            throw new ArgumentNullException();
          else
            hostUri = value;
        }
        else {
          throw new InvalidOperationException("already logged in");
        }
      }
    }

    public string Password {
      get { return password; }
      set
      {
        if (session == null) {
          if (value == null)
            throw new ArgumentNullException();
          else
            password = value;
        }
        else {
          throw new InvalidOperationException("already logged in");
        }
      }
    }

    public bool UseSsl {
      get { return useSsl; }
      set
      {
        if (session == null)
          useSsl = value;
        else
          throw new InvalidOperationException("already logged in");
      }
    }

    public int Timeout {
      get { return timeout; }
      set
      {
        if (timeout != value) {
          timeout = value;
          if (session != null)
            session.Timeout = timeout;
        }
      }
    }

    public PopMailbox Mailbox {
      get { return mailbox; }
      internal set { mailbox = value; }
    }

    internal PopClient(PopClientAgent agent)
    {
      this.agent = agent;
    }

    /*
     * methods for authentication
     */

    /// <summary>tries to login</summary>
    public void Login()
    {
      agent.Connect(this);

      agent.Authenticate(this);
    }

    /// <summary>this method closes selected mailbox if required, and logs out</summary>
    public void Logout()
    {
      agent.Logout(this);
    }

    /*
     * methods for requesting
     */

    /// <summary>lists messages</summary>
    public PopMessage[] ListMessages()
    {
      return agent.ListMessages(this);
    }

    /// <summary>gets message unique-id</summary>
    public string GetUid(PopMessage message)
    {
      return agent.GetUid(this, message);
    }

    /// <summary>updates mailbox status</summary>
    public void UpdateMailboxStatus()
    {
      agent.UpdateMailboxStatus(this);
    }

    /// <summary>retrieves a message header and body from listed message</summary>
    public PopMessage Retrieve(PopMessage message)
    {
      return agent.Retrieve(this, message);
    }

    /// <summary>retrieves a message header and body from listed message. and deletes after retrieving.</summary>
    public PopMessage RetrieveAndDelete(PopMessage message)
    {
      return agent.RetrieveAndDelete(this, message);
    }

    /// <summary>retrieves a message header from listed message</summary>
    public PopMessage RetrieveHeader(PopMessage message)
    {
      return agent.RetrieveHeader(this, message);
    }

    /// <summary>deletes a message</summary>
    public void Delete(PopMessage message)
    {
      agent.Delete(this, message);
    }

    /// <summary>cancels deletion</summary>
    public void CancelDeletion()
    {
      agent.CancelDeletion(this);
    }

    /// <summary>sends NOOP request</summary>
    public void NoOp()
    {
      agent.NoOp(this);
    }

    private readonly PopClientAgent agent;
    private PopSession session = null;
    private PopUri hostUri = null;
    private string password = string.Empty;
    private bool useSsl = false;
    private int timeout = 32 * 1000;
    private PopMailbox mailbox = null;
  }
}
